"""This module contains decorators for using Tremors loggers.

The :func:`logged` decorator wraps a callable in a Tremors logger context,
and injects that logger into the callable.
"""

import functools
import sys
from typing import TYPE_CHECKING, Any, cast, overload

if TYPE_CHECKING:
    from collections.abc import Callable
elif "sphinx" in sys.modules:
    from collections.abc import Callable  # noqa: TC003

import tremors.logger


def _logged[TRet, **P](
    fn: Callable[P, TRet],
    *collectors: tremors.logger.Collector[Any],
    name: str,
    logger_name: str | None = None,
) -> Callable[P, TRet]:
    @functools.wraps(fn)
    def logged_wrapper(*args: P.args, **kwargs: P.kwargs) -> TRet:
        if "logger" in kwargs:
            return fn(*args, **kwargs)
        with tremors.logger.Logger(*collectors, name=name, logger_name=logger_name) as logger:
            kwargs["logger"] = logger
            return fn(*args, **kwargs)

    return logged_wrapper


@overload
def logged[TRet, **P](
    *collectors: tremors.logger.Collector[Any], name: str, logger_name: str | None
) -> Callable[[Callable[P, TRet]], Callable[P, TRet]]: ...


@overload
def logged[TRet, **P](fn_or_collector: Callable[P, TRet]) -> Callable[P, TRet]: ...


def logged[TRet, **P](
    fn_or_collector: Callable[P, TRet] | tremors.logger.Collector[Any] | None = None,
    *collectors: tremors.logger.Collector[Any],
    name: str | None = None,
    logger_name: str | None = None,
) -> Callable[P, TRet] | Callable[[Callable[P, TRet]], Callable[P, TRet]]:
    """Inject a :class:`~tremors.logger.Logger` into a callable.

    The callable must have a keyword-only parameter ``logger``. If the
    value for ``logger`` is the default value, :data:`from_logged`, it is
    replaced by a logger with the specified collectors, name, and underlying
    logger. This logger is entered, and exited immediately before and after
    the callable runs. Otherwise, the supplied logger will be used in the
    call. A supplied logger will not be automatically entered, and exited.

    Args:
        fn_or_collector: If a callable, e.g., the decorator was used like
            ``@logged``, it is assumed that no collectors were specified,
            and the default collectors are used. Otherwise, it is assumed
            this is the first collector, e.g., the decorator was used like
            ``@logged(some_collector, ..., logger_name=...)``.
        *collectors: The rest of the collectors.
        name: The name of the logger. If None, the callable name is used.
        logger_name: The name of the underlying logger. If None, the standard
            root logger is used.
    """
    if callable(fn_or_collector):
        default_collectors = ()  # TODO @NAS: Implement default collectors.
        return _logged(fn_or_collector, *default_collectors, name=fn_or_collector.__name__)

    def decorator(fn: Callable[P, TRet]) -> Callable[P, TRet]:
        return (
            _logged(
                fn,
                fn_or_collector,
                *collectors,
                name=name if name is not None else fn.__name__,
                logger_name=logger_name,
            )
            if fn_or_collector
            else _logged(
                fn,
                *collectors,
                name=name if name is not None else fn.__name__,
                logger_name=logger_name,
            )
        )

    return decorator


from_logged: tremors.logger.Logger = cast("tremors.logger.Logger", None)
"""A sentinel logger value.

When used as the ``logger`` argument to a :func:`logged` callable, the
callable is wrapped in a :class:`~tremors.logger.Logger` context that is
injected into the callable as the ``logger`` argument.
"""



"""This module contains collector bundles and tools.

A collector bundle is an object that contains the definition of
the collector state, a factory function to create a collector, and a
formatter function that will extract, and format the collector state from
a :class:`~logging.LogRecord`.

In addition to the bundles in this module, you can create custom bundles. A
collector is generic in the type of its state, which can be of an arbitrary
type. A collector factory can be defined to have arbitrary parameters that
depend on how the collector is initialized. A collector formatter has a
fixed signature as described in :class:`Formatter`.
"""

import dataclasses
import logging
import sys
import time
from collections.abc import Mapping
from typing import TYPE_CHECKING, NamedTuple, Protocol

import tremors.logger

if TYPE_CHECKING:
    import uuid
    from collections.abc import Callable
elif "sphinx" in sys.modules:
    import uuid  # noqa: TC003
    from collections.abc import Callable


def get_state[T](typ: type[T], name: str, record: logging.LogRecord) -> T | None:
    """Get the collector state from a :class:`~logging.LogRecord`.

    The function can be used by custom collector formatters to get the
    collector state from a record.

    Args:
        typ: The collector state class.
        name: The name of the collector.
        record: The LogRecord the may contain the collector state.

    Returns:
        An object of type ``typ`` if the LogRecord contains such an object
        for the collector. Otherwise None.
    """
    tremors_extra = getattr(record, tremors.logger.EXTRA_KEY, None)
    if isinstance(tremors_extra, Mapping):
        state = tremors_extra.get(name)
        if isinstance(state, typ):
            return state
    return None


class Formatter(Protocol):
    """Callback proctocol for a collector formatter."""

    def __call__(
        self, record: logging.LogRecord, *, name: str = "", fmt: str = "", default: str = ""
    ) -> str:
        """Extract and format a collector state.

        Args:
            record: A :class:`~logging.LogRecord` that may contain the
                collector state.
            name: The name of the collector.
            fmt: A format specification. Implementations commonly expect
                this to adhere to the :ref:`formatspec`. Then ``str(fmt)``
                may be used to obtain the specification, so ``fmt`` does
                not have to be a string.
            default: A value that may be returned if the formatter could
                not obtain, or format the state.

        Returns:
            The state formatted as a string, or the default value.
        """


class CollectorBundle[T, **P](NamedTuple):
    """A bundle to create, and work with a collector."""

    state: type[T]
    factory: Callable[P, tremors.logger.Collector[T]]
    formatter: Formatter


class IndentifierState(NamedTuple):
    """The state of a collector created by :func:`identifier_factory`."""

    group_id: uuid.UUID
    parent: tremors.logger.Logger | None
    path: tuple[str, ...]


def _identifier_collect(
    state: IndentifierState | None, log_item: tremors.logger.LogItem
) -> IndentifierState | None:
    if state:
        return state
    logger = log_item.logger
    if not logger.group_id or not logger.path:
        return None
    return IndentifierState(group_id=logger.group_id, parent=logger.parent, path=logger.path)


def identifier_factory(
    name: str = "identifier", *, level: int = logging.INFO
) -> tremors.logger.Collector[IndentifierState | None]:
    """Create a collector to gather logger identifiers.

    Args:
        name: The collector name.
        level: The collector level.

    Returns:
        An identifier collector with the specified name, and level.
    """
    return tremors.logger.Collector(name=name, level=level, state=None, collect=_identifier_collect)


def identifier_formatter(
    record: logging.LogRecord,
    *,
    name: str = "identifier",
    fmt: str = "[{process}{thread}{group_id}] {path}",
    default: str = "",
) -> str:
    """Extract an :class:`IndentifierState` from a record, and format it.

    ``str(fmt)`` may contain any of the fields in this table. The formatted
    string is generated by replacing the fields with the values described
    in the table.

    ========  =====
    Field     Value
    ========  =====
    process   ``f"p:{pid}"`` where ``pid`` is the process ID if it's
               available (notice the trailing space). Otherwise the empty
               string.
    thread    ``f"t:{tid}"`` where ``tid`` is the thread ID if it's
               available (notice the trailing space). Otherwise the empty
               string.
    group_id  ``f"t:{gid}"`` where ``gid`` is the group ID.
    path      The path represented as a POSIX-like path.
    ========  =====

    Returns:
        The formatted string if the record contains a state for the
        collector. Otherwise ``default``.
    """
    state = get_state(IndentifierState, name, record)

    if state is None:
        return default

    process = f"p:{record.process} " if record.process is not None else ""
    thread = f"t:{record.thread} " if record.thread is not None else ""
    group_id = f"g:{state.group_id}"
    path = "/".join(state.path)
    return str(fmt).format(process=process, thread=thread, group_id=group_id, path=path)


identifier = CollectorBundle(
    state=IndentifierState, factory=identifier_factory, formatter=identifier_formatter
)
"""The identifier collector bundle.

Examples:

    With :func:`~logging.basicConfig` we configure the standard root logger
    with a stream handler that has a formatter that incorporates the record's
    ``ident`` attribute in the message. We add the filter function ``flt`` to
    the handler. The filter uses the formatter from the identifier collector
    bundle to extract the collector state from a record, and format it.
    Then the filter returns a modified record with the ident attribute set
    to the formatted state.

    .. note::

        If we had added the filter to the logger instead of the handler,
        the filter would not be called for messages logged by descendant
        standard loggers.  But by adding the filter to the handler, we ensure
        that it is called for all messages that the handler could emit.

    .. code-block:: python

        import copy
        import logging

        import tremors
        from tremors import collector


        def flt(record: logging.LogRecord) -> logging.LogRecord:
            record = copy.copy(record)
            ident = collector.identifier.formatter(record)
            record.ident = f"{ident} > " if ident else ""
            return record


        @tremors.logged(collector.identifier.factory())
        def parent(*, logger: tremors.Logger = tremors.from_logged) -> None:
            logger.info("foo")
            child()


        @tremors.logged(collector.identifier.factory())
        def child(*, logger: tremors.Logger = tremors.from_logged) -> None:
            logger.info("bar")


        logging.basicConfig(
            format="%(ident)s%(levelname)s:%(name)s:%(message)s", level=logging.INFO
        )
        logging.root.handlers[0].addFilter(flt)
        parent()

    The logged messages contain the process ID, thread ID, group ID, and
    hierarchy as a path.

    .. code-block:: shell

        [p:... t:... g:...] parent > INFO:root:entered: parent
        [p:... t:... g:...] parent > INFO:root:foo
        [p:... t:... g:...] parent/child > INFO:root:entered: child
        [p:... t:... g:...] parent/child > INFO:root:bar
        [p:... t:... g:...] parent/child > INFO:root:exited: child
        [p:... t:... g:...] parent > INFO:root:exited: parent

    We can expand on the previous example by throwing a function that uses
    standard logging into the mix. We modify ``child`` to call ``other``,
    which could be a function that you can't change, and doesn't use Tremors
    loggers. By default, standard loggers propagate their messages, so
    descendant messages will eventually be processed by the root logger's
    handler with our filter. The collector formatter will not be able to
    extract the state from a descendant record, and will return the default
    value, the empty string. Then ``record.ident`` will be the empty string,
    which will be interpolated into the formatted message.

    .. note::

        Since the formatter expects the record to have an ident attribute,
        the filter must always add this attribute to the record.

    .. code-block:: python

        @tremors.logged(collector.identifier.factory())
        def child(*, logger: tremors.Logger = tremors.from_logged) -> None:
            logger.info("bar")
            other()


        def other() -> None:
            logging.getLogger(__name__).info("baz")


        parent()

    The functions that use Tremors loggers have entering and exiting
    messages, and their messages contain identifier information, as before.
    The ``other`` function, which logs with standard loggers simply emits
    messages with empty ident sections.

    .. code-block:: shell

        [p:... t:... g:...] parent > INFO:root:entered: parent
        [p:... t:... g:...] parent > INFO:root:foo
        [p:... t:... g:...] parent/child > INFO:root:entered: child
        [p:... t:... g:...] parent/child > INFO:root:bar
        INFO:__main__:baz
        [p:... t:... g:...] parent/child > INFO:root:exited: child
        [p:... t:... g:...] parent > INFO:root:exited: parent
"""


@dataclasses.dataclass
class CounterState:
    """The state of a collector created by :func:`identifier_factory`.

    This state is mutable so that a single collector instance can be shared
    with multiple loggers, and each logger can update the same state.
    """

    count: int
    step: int


def _counter_collect(state: CounterState, _: tremors.logger.LogItem) -> CounterState:
    state.count += state.step
    return state


def counter_factory(
    name: str = "counter", *, level: int = logging.INFO, initial: int = 0, step: int = 1
) -> tremors.logger.Collector[CounterState]:
    """Create a collector to count logging calls.

    Args:
        name: The collector name.
        level: The collector level.
        initial: The initial value of the count.
        step: How much to increase the count by each time the collector runs.

    Returns:
        A counter collector with the specified name, level, behavior, and mutable state.
    """
    return tremors.logger.Collector(
        name=name,
        level=level,
        state=CounterState(count=initial, step=step),
        collect=_counter_collect,
    )


def counter_formatter(
    record: logging.LogRecord, *, name: str = "counter", fmt: str = "{counter}", default: str = "0"
) -> str:
    """Extract a count from a record, and format it.

    ``str(fmt)`` may contain the field ``counter``. The formatted string is
    generated by replacing the field with the state's count.

    Returns:
        The formatted string if the record contains a state for the
        collector. Otherwise ``default``.
    """
    state = get_state(CounterState, name, record)

    if state is None:
        return default

    return str(fmt).format(counter=state.count)


counter = CollectorBundle(state=CounterState, factory=counter_factory, formatter=counter_formatter)
"""The counter collector bundle.

Examples:

    We configure standard logging in a similar manner to the examples in
    :attr:`identifier`, except here we use counter collectors.

    .. code-block:: python

        import copy
        import logging

        import tremors
        from tremors import collector


        def flt(record: logging.LogRecord) -> logging.LogRecord:
            record = copy.copy(record)
            count = collector.counter.formatter(record, fmt="[{counter}] ", default="")
            record.count = count
            return record


        @tremors.logged(collector.counter.factory())
        def parent(*, logger: tremors.Logger = tremors.from_logged) -> None:
            logger.info("foo")
            child()


        @tremors.logged(collector.counter.factory())
        def child(*, logger: tremors.Logger = tremors.from_logged) -> None:
            logger.info("bar")


        logging.basicConfig(
            format="%(count)s%(levelname)s:%(name)s:%(message)s", level=logging.INFO
        )
        logging.root.handlers[0].addFilter(flt)
        parent()

    The logged messages include the count, where the ``parent`` and ``child``
    counts are distinct from each other because each loggers uses its own
    counter collector.

    .. code-block:: shell

        [1] INFO:root:entered: parent
        [2] INFO:root:foo
        [1] INFO:root:entered: child
        [2] INFO:root:bar
        [3] INFO:root:exited: child
        [3] INFO:root:exited: parent

    Now let's modify the previous example to have ``parent`` and ``child``
    use a shared counter.

    .. code-block:: python

        shared_counter = collector.counter.factory()


        @tremors.logged(shared_counter)
        def parent(*, logger: tremors.Logger = tremors.from_logged) -> None:
            logger.info("foo")
            child()


        @tremors.logged(shared_counter)
        def child(*, logger: tremors.Logger = tremors.from_logged) -> None:
            logger.info("bar")


        parent()

    Here the count is shared between the two functions.

    .. code-block:: shell

        [1] INFO:root:entered: parent
        [2] INFO:root:foo
        [3] INFO:root:entered: child
        [4] INFO:root:bar
        [5] INFO:root:exited: child
        [6] INFO:root:exited: parent
"""


@dataclasses.dataclass
class ElapsedState:
    """The state of an :func:`elapsed` collector.

    The state is mutable like :class:`CounterState`.
    """

    t0: int | None
    t: int | None


def _collect_elapsed(state: ElapsedState, _: tremors.logger.LogItem) -> ElapsedState:
    t = time.perf_counter_ns()
    if state.t0 is None:
        state.t0 = t
    state.t = t
    return state


def elapsed_factory(
    name: str = "elapsed", *, level: int = logging.INFO
) -> tremors.logger.Collector[ElapsedState]:
    """Create a collector to measure how much time has elapsed.

    Args:
        name: The collector name.
        level: The collector level.

    Returns:
        An elapsed collector with the specified name, and level.
    """
    return tremors.logger.Collector(
        name=name, level=level, state=ElapsedState(t0=None, t=None), collect=_collect_elapsed
    )


def elapsed_formatter(
    record: logging.LogRecord,
    *,
    name: str = "elapsed",
    fmt: str = "{elapsed:.3f}",
    default: str = "",
) -> str:
    """Extract an :class:`ElapsedState` from a record, and format it.

    ``str(fmt)`` may contain the field ``elapsed``. The formatted string is
    generated by replacing the field with the state's elapsed time in seconds.

    Returns:
        The formatted string if the record contains a state for the
        collector. Otherwise ``default``.
    """
    state = get_state(ElapsedState, name, record)

    if state is None:
        return default

    t0, t = state.t0, state.t

    if t0 is None or t is None:
        return default

    elapsed_s = (t - t0) / 1e9
    return str(fmt).format(elapsed=elapsed_s)


elapsed = CollectorBundle(state=ElapsedState, factory=elapsed_factory, formatter=elapsed_formatter)
"""The elapsed collector bundle.

Examples:

    We configure standard logging in a similar manner to the examples in
    :attr:`identifier`, except here we use elapsed collectors, and identifier
    collectors with a custom format to only show the path.

    .. code-block:: python

        import copy
        import logging
        import time

        import tremors
        from tremors import collector


        def flt(record: logging.LogRecord) -> logging.LogRecord:
            record = copy.copy(record)
            ident = collector.identifier.formatter(record, fmt="{path}")
            elapsed = collector.elapsed.formatter(record)
            record.ident = f"{ident} > " if ident else ""
            record.elapsed = f"{elapsed} " if elapsed else ""
            return record


        @tremors.logged(collector.identifier.factory(), collector.elapsed.factory())
        def parent(*, logger: tremors.Logger = tremors.from_logged) -> None:
            logger.info("sleeping for 1s...")
            time.sleep(1)
            child()


        @tremors.logged(collector.identifier.factory(), collector.elapsed.factory())
        def child(*, logger: tremors.Logger = tremors.from_logged) -> None:
            logger.info("sleeping for 1s...")
            time.sleep(1)


        logging.basicConfig(
            format="%(elapsed)s%(ident)s%(levelname)s:%(name)s:%(message)s",
            level=logging.INFO,
        )
        logging.root.handlers[0].addFilter(flt)
        parent()

    The logged messages include the elapsed time since each elapsed collector
    started.

    .. code-block:: shell

        0.000 parent > INFO:root:entered: parent
        0.000 parent > INFO:root:sleeping for 1s...
        0.000 parent/child > INFO:root:entered: child
        0.000 parent/child > INFO:root:sleeping for 1s...
        1.001 parent/child > INFO:root:exited: child
        2.001 parent > INFO:root:exited: parent

    We can try the same thing again, only this time with a shared elapsed
    collector.

    .. code-block:: python

        shared_elapsed = collector.elapsed.factory()


        @tremors.logged(collector.identifier.factory(), shared_elapsed)
        def parent(*, logger: tremors.Logger = tremors.from_logged) -> None:
            logger.info("sleeping for 1s...")
            time.sleep(1)
            child()


        @tremors.logged(collector.identifier.factory(), shared_elapsed)
        def child(*, logger: tremors.Logger = tremors.from_logged) -> None:
            logger.info("sleeping for 1s...")
            time.sleep(1)


        parent()

    Notice that this time, the ``child`` elapsed time starts at 1s instead
    of resetting to 0s.

    .. code-block:: shell

        0.000 parent > INFO:root:entered: parent
        0.000 parent > INFO:root:sleeping for 1s...
        1.000 parent/child > INFO:root:entered: child
        1.001 parent/child > INFO:root:sleeping for 1s...
        2.001 parent/child > INFO:root:exited: child
        2.001 parent > INFO:root:exited: parent
"""



"""This module contains the Tremors Logger and Collector definition.

A Tremors logger can be used as a drop-in replacement for a standard
:class:`~logging.Logger`. Collectors may be attached to a logger. Each time
a message is logged, the collectors run if they are enabled for the message
level. When a collector runs, it may update its state. A logger adds the
current state of all of its collectors to any :class:`~logging.LogRecord`
that it produces.

A logger may also be used as a context manager. When used this way, loggers
maintain a hierarchy. Each logger has a parent logger, unless it is the root
logger. The root logger assigns a :attr:`~Logger.group_id` that is shared
by all loggers in the hierarchy. Each logger has a :attr:`~Logger.path` from
its root logger to itself. Hierarchies may be nested by explicitly creating
a new root logger within an existing hierarchy.  The loggers in the nested
hierarchy will have a different group ID than the loggers in the containing
hierarchy. The paths in the nested hierarchy will also start from the new
root logger.

When a logger context is entered, or exited, a message is automatically
logged at a specified level, and collectors will run if they are enabled
for the message level. This allows us to define collectors that measure
lifecycle information, such as the duration of a context, or how much memory
was allocated at the beginning, and end of a context.

Each Tremors logger uses an underlying standard logger that may
be specified. The underlying logger can be configured in the normal
fashion. Underlying loggers may be shared by Tremors loggers. In fact,
it is common for all Tremors loggers to use the standard root logger. The
Tremors logger simply adds the states of its collectors to the records
produced by its standard logger. These augmented records can then be
processed by any :class:`~logging.Filter`, :class:`~logging.Formatter`,
or :class:`~logging.Handler` that knows about the extra record attributes.

A collector may be bundled with a formatter that can extract the collector
state from a record, and format it. This formatter can be used in a
filter, formatter, or handler that has been configured for the underling
logger. Tremors comes with many useful collector bundles. But you can also
define custom collectors, and bundles.
"""

from __future__ import annotations

import contextvars
import functools
import itertools
import logging
import sys
import uuid
from typing import TYPE_CHECKING, Any, NamedTuple, Self

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping, MutableMapping
    from types import TracebackType
elif "sphinx" in sys.modules:
    from collections.abc import Callable, Mapping  # noqa: TC003
    from types import TracebackType  # noqa: TC003

_current_logger: contextvars.ContextVar[Logger | None] = contextvars.ContextVar(
    "_current_logger", default=None
)


class LogItem(NamedTuple):
    """The properties of a logged message.

    This is essentially the arguments passed to :meth:`Logger.log`.
    """

    logger: Logger
    level: int
    msg: object
    args: tuple[object, ...]
    exc_info: (
        bool
        | tuple[type[BaseException], BaseException, TracebackType | None]
        | tuple[None, None, None]
        | BaseException
        | None
    )
    stack_info: bool
    stacklevel: int
    extra: Mapping[str, object] | None
    kwargs: dict[str, object]


class Collector[TState](NamedTuple):
    """The collector specification.

    Attributes:
        name: The name of the collector. When the collector runs
            for a logged message, the collector's state is added to the
            :class:`~logging.LogRecord`, and may be retrieved using this name.
        level: The minimum level a logged message must be for the collector
            to run.
        state: The initial state of the collector. Each time the collector
            runs, it may update this state.
        collect: When the collector runs, this function is called with the
            current state, and the LogItem. It returns the new state.
    """

    name: str
    level: int
    state: TState
    collect: Callable[[TState, LogItem], TState]


def _collectors_reducer[T](
    acc: tuple[MutableMapping[str, Collector[T]], MutableMapping[str, T]],
    curr: tuple[LogItem, tuple[str, Collector[T]]],
) -> tuple[MutableMapping[str, Collector[T]], MutableMapping[str, T]]:
    acc_collectors, acc_extra = acc
    log_item, (name, curr_c) = curr
    if curr_c.level != logging.NOTSET and log_item.level < curr_c.level:
        acc_collectors[name] = curr_c
        acc_extra[name] = curr_c.state
        return acc_collectors, acc_extra
    state = curr_c.collect(curr_c.state, log_item)
    acc_collectors[name] = Collector(curr_c.name, curr_c.level, state, curr_c.collect)
    acc_extra[name] = state
    return acc_collectors, acc_extra


EXTRA_KEY = "tremors"


class Logger(logging.LoggerAdapter[logging.Logger]):
    """The Tremors logger.

    Args:
        *collectors: Collectors that will be attached to the logger, and will
            be run for every logged message for which the collector is enabled.
        name: The name of the logger. This name is logged in entering, and
            exiting messages. This name is used to generate the path, but
            may be altered in the path to be unique in the hierarchy.
        logger_name: The name of the underlying :class:`~logging.Logger`
            that will be used to log messages. If None, the root logger will
            be used.
        ctx_level: The level at which entered and exited messages will
            be logged.
        is_root: If True, a root logger with no parent that starts a new
            hierarchy will be created.
    """

    def __init__(
        self,
        *collectors: Collector[Any],
        name: str,
        logger_name: str | None = None,
        ctx_level: int = logging.INFO,
        is_root: bool = False,
    ) -> None:
        """Initialize the logger."""
        super().__init__(logging.getLogger(logger_name))
        self._name = name
        self._collectors: Mapping[str, Collector[Any]] = {c.name: c for c in collectors}
        self._entered = 0
        self._ctx_level = ctx_level
        self._cv_token: contextvars.Token[Logger | None] | None = None
        self._is_root = is_root
        self._parent: Logger | None = None
        self._group_id: uuid.UUID | None = None
        self._path: tuple[str, ...] | None = None
        self._path_registry: dict[str, int] = {}

    def __enter__(self) -> Self:
        """Enter the context.

        The logger will only be added to the hierarchy, and log the entered
        message if the logger is not currently entered, i.e., it has not
        been entered before, or the number of times it has been entered,
        and exited are balanced.

        If the logger is added to the hierarchy, an entered message will be
        logged at the ``ctx_level`` the logger was initialized with. These
        messages may be suppressed by specifying a sufficiently low
        level. These messages may also be filtered out. The message text
        will be ``"entered: %s" % name`` where ``name`` is the name the
        logger was initialized with.
        """
        self._entered += 1
        if self._entered != 1:
            return self
        if not self._is_root:
            self._parent = _current_logger.get()
        if self._parent:
            self._group_id = self._parent.group_id
            self._path = self._parent.register_path(self)
        else:
            self._group_id = uuid.uuid4()
            self._path = (self._name,)
        self._cv_token = _current_logger.set(self)
        self.log(self._ctx_level, "entered: %s", self._name)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the context.

        The logger will only be removed from the hierarchy if it is exiting
        for the last time, i.e., the number of times it has been entered,
        and exited are balanced.

        If the logger is removed from the hierarchy, or it is exiting
        due to an exception, an exited message will be logged. See
        :meth:`~Logger.__enter__` for how the message is logged, and how to
        suppress it.  The message text will be ``"exited: %s" % name``.
        """
        self._entered -= 1
        exc_info = (exc_type, exc_val, exc_tb) if exc_type and exc_val else None
        if self._entered == 0 or exc_info:
            self.log(self._ctx_level, "exited: %s", self._name, exc_info=exc_info)
        if self._entered == 0 and self._cv_token:
            _current_logger.reset(self._cv_token)

    @property
    def group_id(self) -> uuid.UUID | None:
        """The group ID for the hierarchy assigned by the root logger."""
        return self._group_id

    @property
    def name(self) -> str:
        """The logger name.

        This is not the name of the underlying Python logger. This is the name
        of the logger in the path if the logger has been entered. Otherwise,
        it is the ``name`` that the logger was initialized with.
        """
        return self._path[-1] if self._path else self._name

    @property
    def parent(self) -> Logger | None:
        """The parent logger, or None if this is a root logger."""
        return self._parent

    @property
    def path(self) -> tuple[str, ...] | None:
        """The logger path.

        The path is a sequence of logger names following the hierarchy of
        loggers from the root logger to this one. The final name in the
        path will be the name that the logger was initialized with if a
        logger with the same name has not been registered at that path index
        yet. Otherwise, an incremental number is appended to the initial
        name to get a unique name for that path index.

        Example:
            If there are 2 loggers named "baz" at the third level of the
            hierarchy, their respecitve paths will be:

            .. code-block:: python

                ["foo", "bar", "baz"]
                ["foo", "bar", "baz2"]
        """
        return self._path

    def register_path(self, logger: Logger) -> tuple[str, ...]:
        """A logger can register a path for itself with its parent.

        This method should not be called directly. The logger will
        automatically register itself with its parent.
        """
        if logger.parent is not self:
            msg = "logger is not a child of this logger."
            raise RuntimeError(msg)
        if not self._path:
            msg = "path has not been initialized"
            raise RuntimeError(msg)
        name = logger.name
        if name not in self._path_registry:
            self._path_registry[name] = 1
            return (*self._path, name)
        self._path_registry[name] += 1
        return (*self._path, f"{name}{self._path_registry[name]}")

    def log(  # noqa: PLR0913
        self,
        level: int,
        msg: object,
        *args: object,
        exc_info: bool
        | tuple[type[BaseException], BaseException, TracebackType | None]
        | tuple[None, None, None]
        | BaseException
        | None = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Mapping[str, object] | None = None,
        **kwargs: object,
    ) -> None:
        """Run collectors, and log a message at the given level.

        The arguments are interpreted as for :meth:`~logging.Logger.debug`.
        """
        initial_collectors: MutableMapping[str, Collector[Any]] = {}
        initial_extra: MutableMapping[str, Any] = {}
        log_item = LogItem(
            logger=self,
            level=level,
            msg=msg,
            args=args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
            kwargs=kwargs,
        )
        self._collectors, own_extra = functools.reduce(
            _collectors_reducer,
            zip(itertools.repeat(log_item), self._collectors.items(), strict=False),
            (initial_collectors, initial_extra),
        )
        return self.logger.log(
            level,
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra={**extra, EXTRA_KEY: own_extra} if extra is not None else {EXTRA_KEY: own_extra},
            **kwargs,
        )



"""The tremors package."""

from tremors.decorator import from_logged, logged
from tremors.logger import EXTRA_KEY, Collector, Logger

__authors__ = ["Narvin Singh"]
__project__ = "Tremors"
__version__ = "0.4.0"

__all__ = ["EXTRA_KEY", "Collector", "Logger", "__version__", "from_logged", "logged"]
